// Fitting 
// A <=> B  
//
#include <boost/array.hpp>
#include <boost/numeric/odeint.hpp>
#include <vector>
#include <iomanip>
#include <fstream>
#include <vector>
#include <cmath>
#include <iostream>
#include <sstream>
#include "Read.h" 
#include "Stats.h" 
#include <dlib/optimization.h> 

using namespace std;
using namespace boost::numeric::odeint;

typedef dlib::matrix<double,3,1> state_type;		// define initial state type vector with the number of peaks + 1 time column 
typedef dlib::matrix<double,2,1> params_type;		// define params type vector with the number of rate constants

  ofstream file3("sim.txt");				// output file "sim.txt" for simulated fits

// ----------------------------------------------------------------------------------------
// we're using the ODEINT ordinary diffential equation integrator supplied within boost

struct ode4
{
  double k1  = 0.40 ; 
  double k1r = 0.50 ; 

  ode4( 
	double m_k1  = 0.50 , 
	double m_k1r = 0.50  
	)
    	: k1(m_k1) , k1r(m_k1r) { }

        void operator()( const state_type &x , state_type &dxdt , double t )
        {

	// here are the diffeqs from the rxn mechanism:
	
  	dxdt(0) = -k1 * x(0)  + k1r * x(1) ;
  	dxdt(1) =  k1 * x(0)  - k1r * x(1) ;

  }
} ;

// ----------------------------------------------------------------------------------------

double rateCurve( const params_type& params )
{
  const char* FILE_NAME = "dat.txt";		// read in raw data file
  std::vector< std::vector<double> > A = fileIn( FILE_NAME );

  double t_dw = 0.01       ;
  double k1  = params(0,0) ;
  double k1r = params(0,1) ;

  double distances = 0 ;        		// residuals sum
  typedef runge_kutta4< state_type > stepper_type;

  for( long unsigned int i = 0 ; i < A.size() ; i++ )    {

    state_type x = { A[0][1] , A[0][2] } ; 	// initial conditions 

    integrate_const( stepper_type() , ode4( k1 , k1r ) , x , A[0][0] , A[i][0] , t_dw );
    distances = dist( A[i][1] ,  x(0) ) +	// residual btw first peak and its simulated value 
		dist( A[i][2] ,  x(1) ) + 	// residual btw second peak and its simulated value 
		distances ;
  }
  return distances ;	// return total residual value btw all peaks and simulated data / this is what gets minimized during fitting
}

// ----------------------------------------------------------------------------------------

void write_fit( const state_type &x , const double t )
{
  double a = t * 10 ;
  double b = floor(t * 10) ;            // floor statement only to print 0.1 units of time increments
  if (abs(b-a) < 0.001 ){               // statement to correct for rounding errors b4 proceeding 
  file3 << fixed << setw(8) << setprecision(6) <<  t 
 		 << ' ' << x(0)  << ' ' << x(1)         
      		 << endl ;       
  }
}

// ----------------------------------------------------------------------------------------

int main()
{

  const char* FILE_NAME = "dat.txt";					// define data file to read-in
  std::vector< std::vector<double> > A = fileIn( FILE_NAME );		// push data file into matrix A / see Read.h 
  cout << endl << " Reading \"dat.txt\" file. " << endl << endl ;

  const char* FILE_NAMEB = "guess.txt";					// define file with parameters to read-in
  std::vector< std::vector<double> > B = fileIn( FILE_NAMEB );		// push file into matrix A / see Read.h 
  cout << " Reading \"guess.txt\" file for initial rate constants. " << endl << endl ;

  dlib::matrix<double,2,1> k = vec2matrix( B ) ;			// convert rate constant vector to dlib format

// ----------------------------------------------------------------------------------------
// this is the fitting routine supplied by dlib / there are many to choose 

  dlib::find_min_box_constrained(dlib::bfgs_search_strategy(),  
                             dlib::objective_delta_stop_strategy(1e-9),  
                             rateCurve, derivative(rateCurve), k, 0.001 , 20.0 );

// ----------------------------------------------------------------------------------------
// print out simulated rates into file "results.txt" 

  ofstream file2;
  file2.open("results.txt");
  file2 << fixed << setw(0) << setprecision(3);
  file2 << " k1  = " << k(0,0) << endl  
	<< " k1r = " << k(0,1) << endl	;    
  file2.close();

// ----------------------------------------------------------------------------------------
// pass simulated values and regenerate curves

  double t_init = 	A[0][0]  	;		// initial time from raw data 
  double t_dw =   	0.01       	;		// dwell time between runge-kutta points / steps
  auto i  = 		A.size()   	;		// total number of rows in matrix A/data
  double t_final = 	A[i-1][0]  	;		// last time point in A

  state_type x = { A[0][1] , A[0][2] } ; 		// initial conditions from raw data first row

  typedef runge_kutta4< state_type > stepper_type;	// define runge-kutta stepper for numerical integration
  integrate_const( stepper_type() , ode4( k(0,0) , k(0,1) ) , x , t_init , t_final, t_dw , write_fit ) ;

// ----------------------------------------------------------------------------------------

  cout << " Saving \"results.txt\" file with optimal rate constants. " << endl << endl ;
  cout << " Saving \"sim.txt\" file for simulation curves. " << endl << endl ;
  cout 	<< fixed << setw(0) << setprecision(3);
  cout 	<< " k1  = " << k(0,0) << endl  
	<< " k1r = " << k(0,1) << endl << endl ; 

// ----------------------------------------------------------------------------------------
// calc RMSE and residuals 

  const char* FILE_NAMEC = "sim.txt";                                   // define data file to read-in
  std::vector< std::vector<double> > C = fileIn( FILE_NAMEC );          // push data file into matrix A / see Read.h 
  cout << endl << " Reading \"sim.txt\" file. " << endl << endl ;

  print_residuals ( A , C ) ;

  file2.close();
  file3.close();

  return 0;

}

// ----------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------
